"""
Utils module.

This module contains simple utility classes and functions.
"""
import signal

import pkg_resources
from loguru import logger


class SignalHandler:
    """A helper class to handle signals."""

    def __init__(self, signals):
        """
        Initialization method.

        Args:
            signals (list of str): List of signals names as found in the ``signal`` module (example: SIGTERM).
        """
        logger.debug("Signal handler: handling signals " + ", ".join(signals))
        self.triggered = False
        for sig in signals:
            try:
                signal.signal(signal.Signals[sig], self.trigger)
            except ValueError as error:
                logger.error(f"Failed to setup signal handler for {sig}: {error}")

    def __bool__(self):
        """Return True when one of the given signal was received, False otherwise."""
        return self.triggered

    # pylint: disable=unused-argument
    def trigger(self, signum, frame):
        """Mark this instance as 'triggered' (a specified signal was received)."""
        logger.debug(f"Signal handler: caught signal {signal.Signals(signum).name} ({signum})")
        self.triggered = True


class Version:
    def __init__(self, version_string):
        self.version_string = version_string
        self.major, self.minor, self.patch = (int(v) for v in version_string.split("."))

    def __str__(self):
        return self.version_string


def human_readable_timedelta(value, precision=0):
    """Return a human-readable time delta as a string."""
    pieces = []

    if value.days:
        pieces.append(f"{value.days}d")

    seconds = value.seconds

    if seconds >= 3600:
        hours = int(seconds / 3600)
        pieces.append(f"{hours}h")
        seconds -= hours * 3600

    if seconds >= 60:
        minutes = int(seconds / 60)
        pieces.append(f"{minutes}m")
        seconds -= minutes * 60

    if seconds > 0 or not pieces:
        pieces.append(f"{seconds}s")

    if not precision:
        return "".join(pieces)

    return "".join(pieces[:precision])


def human_readable_bytes(value, digits=2, delim="", postfix=""):
    """
    Return a human-readable bytes value as a string.

    Args:
        value (int): the bytes value.
        digits (int): how many decimal digits to use.
        delim (str): string to add between value and unit.
        postfix (str): string to add at the end.

    Returns:
        str: the human-readable version of the bytes.
    """
    chosen_unit = "B"
    for unit in ("KiB", "MiB", "GiB", "TiB"):
        if value > 1000:
            value /= 1024
            chosen_unit = unit
        else:
            break
    return f"{value:.{digits}f}" + delim + chosen_unit + postfix


def bool_or_value(value):
    """Return True for 'true', False for 'false', original value otherwise."""
    if value == "true":
        return True
    if value == "false":
        return False
    return value


def bool_to_str(value):
    """Return 'true' for True, 'false' for False, original value otherwise."""
    if value is True:
        return "true"
    if value is False:
        return "false"
    return value


def get_version():
    try:
        distribution = pkg_resources.get_distribution("aria2p")
    except pkg_resources.DistributionNotFound:
        return Version("0.0.0")
    else:
        return Version(distribution.version)



"""
This module defines the API class, which makes use of a JSON-RPC client to provide higher-level methods to
interact easily with a remote aria2c process.
"""
import shutil
import threading
from base64 import b64encode
from pathlib import Path

from loguru import logger

from .client import Client, ClientException
from .downloads import Download
from .options import Options
from .stats import Stats
from .utils import get_version


class API:
    """
    A class providing high-level methods to interact with a remote aria2c process.

    This class is instantiated with a reference to a :class:`~aria2p.client.Client` instance. It then uses this client
    to call remote procedures, or remote methods. While the client methods reflect exactly what aria2c is providing
    through JSON-RPC, this class's methods allow for easier / faster control of the remote process. It also
    wraps the information the client retrieves in Python object, like  :class:`~aria2p.downloads.Download`, allowing for
    even more Pythonic interactions, without worrying about payloads, responses, JSON, etc..
    """

    def __init__(self, client=None):
        """
        Initialization method.

        Args:
            client (:class:`~aria2p.client.Client`): an instance of the ``Client`` class.
        """
        if client is None:
            client = Client()
        self.client = client
        self.listener = None

    def __repr__(self):
        return f"API({self.client!r})"

    def add_magnet(self, magnet_uri, options=None, position=None):
        """
        Add a download with a Magnet URI.

        Args:
            magnet_uri (str): the Magnet URI.
            options (:class:`~aria2p.options.Options` or dict): an instance of the ``Options`` class or a dictionary
              containing aria2c options to create the download with.
            position (int): the position where to insert the new download in the queue. Start at 0 (top).

        Returns:
            :class:`~aria2p.downloads.Download` instance: the newly created download object.
        """
        if options is None:
            options = {}

        if isinstance(options, Options):
            client_options = options.get_struct()
        else:
            client_options = options

        gid = self.client.add_uri([magnet_uri], client_options, position)

        return self.get_download(gid)

    def add_torrent(self, torrent_file_path, uris=None, options=None, position=None):
        """
        Add a download with a torrent file (usually .torrent extension).

        Args:
            torrent_file_path (str/Path): the path to the torrent file.
            uris (list of str): a list of URIs used for Web-seeding.
            options (:class:`~aria2p.options.Options` or dict): an instance of the ``Options`` class or a dictionary
              containing aria2c options to create the download with.
            position (int): the position where to insert the new download in the queue. Start at 0 (top).

        Returns:
            :class:`~aria2p.downloads.Download` instance: the newly created download object.
        """
        if uris is None:
            uris = []

        if options is None:
            options = {}

        if isinstance(options, Options):
            client_options = options.get_struct()
        else:
            client_options = options

        with open(torrent_file_path, "rb") as stream:
            torrent_contents = stream.read()
        encoded_contents = b64encode(torrent_contents).decode("utf8")

        gid = self.client.add_torrent(encoded_contents, uris, client_options, position)

        return self.get_download(gid)

    def add_metalink(self, metalink_file_path, options=None, position=None):
        """
        Add a download with a Metalink file.

        Args:
            metalink_file_path (str/Path): the path to the Metalink file.
            options (:class:`~aria2p.options.Options` or dict): an instance of the ``Options`` class or a dictionary
              containing aria2c options to create the download with.
            position (int): the position where to insert the new download in the queue. Start at 0 (top).

        Returns:
            list of :class:`~aria2p.downloads.Download`: the newly created download objects.
        """
        if options is None:
            options = {}

        if isinstance(options, Options):
            client_options = options.get_struct()
        else:
            client_options = options

        with open(metalink_file_path, "rb") as stream:
            metalink_contents = stream.read()
        encoded_contents = b64encode(metalink_contents).decode("utf8")

        gids = self.client.add_metalink(encoded_contents, client_options, position)

        return self.get_downloads(gids)

    def add_uris(self, uris, options=None, position=None):
        """
        Add a download with a URL (or more).

        Args:
            uris (list of str): a list of URIs that point to the same resource.
            options (:class:`~aria2p.options.Options` or dict): an instance of the ``Options`` class or a dictionary
              containing aria2c options to create the download with.
            position (int): the position where to insert the new download in the queue. Start at 0 (top).

        Returns:
            :class:`~aria2p.downloads.Download` instance: the newly created download object.

        """
        if options is None:
            options = {}

        if isinstance(options, Options):
            client_options = options.get_struct()
        else:
            client_options = options

        gid = self.client.add_uri(uris, client_options, position)

        return self.get_download(gid)

    def search(self, patterns):
        """
        Not implemented.

        Search and return :class:`~aria2p.downloads.Download` object based on multiple patterns.

        Args:
            patterns (list of dict): the patterns used to filter the download list.

        Returns:
            list of :class:`~aria2p.downloads.Download` instances: the download objects matching the patterns.

        """
        # gid
        # status
        # totalLength
        # completedLength
        # uploadLength
        # bitfield
        # downloadSpeed
        # uploadSpeed
        # infoHash
        # numSeeders
        # seeder
        # pieceLength
        # numPieces
        # connections
        # errorCode
        # errorMessage
        # followedBy
        # following
        # belongsTo
        # dir
        # files
        # bittorrent
        #        announceList
        #        comment
        #        creationDate
        #        mode
        #        info
        #               name
        # verifiedLength
        # verifyIntegrityPending
        raise NotImplementedError

    def get_download(self, gid):
        """
        Get a :class:`~aria2p.downloads.Download` object thanks to its GID.

        Args:
            gid (str): the GID of the download to get.

        Returns:
            :class:`~aria2p.downloads.Download` instance: the retrieved download object.
        """
        return Download(self, self.client.tell_status(gid))

    def get_downloads(self, gids=None):
        """
        Get a list of :class:`~aria2p.downloads.Download` object thanks to their GIDs.

        Args:
            gids (list of str): the GIDs of the downloads to get. If None, return all the downloads.

        Returns:
            list of :class:`~aria2p.downloads.Download` instances: the retrieved download objects.
        """
        downloads = []

        if gids:
            for gid in gids:
                downloads.append(Download(self, self.client.tell_status(gid)))
        else:
            structs = []
            structs.extend(self.client.tell_active())
            structs.extend(self.client.tell_waiting(0, 1000))
            structs.extend(self.client.tell_stopped(0, 1000))
            downloads = [Download(self, struct) for struct in structs]

        return downloads

    def move(self, download, pos):
        """
        Move a download in the queue, relatively to its current position.

        Args:
            download (:class:`~aria2p.downloads.Download`): the download object to move.
            pos (int): the relative position (1 to move down, -1 to move up, -2 to move up two times, etc.).

        Returns:
            int: The new position of the download.
        """
        return self.client.change_position(download.gid, pos, "POS_CUR")

    def move_to(self, download, pos):
        """
        Move a download in the queue, with absolute positioning.

        Args:
            download (:class:`~aria2p.downloads.Download`): the download object to move.
            pos (int): the absolute position in the queue where to move the download. 0 for top, -1 for bottom.

        Returns:
            int: The new position of the download.
        """
        if pos < 0:
            how = "POS_END"
            pos = -pos
        else:
            how = "POS_SET"
        return self.client.change_position(download.gid, pos, how)

    def move_up(self, download, pos=1):
        """
        Move a download up in the queue.

        Args:
            download (:class:`~aria2p.downloads.Download`): the download object to move.
            pos (int): number of times to move up. With negative values, will move down (use move or move_down instead).

        Returns:
            int: The new position of the download.
        """
        return self.client.change_position(download.gid, -pos, "POS_CUR")

    def move_down(self, download, pos=1):
        """
        Move a download down in the queue.

        Args:
            download (:class:`~aria2p.downloads.Download`): the download object to move.
            pos (int): number of times to move down. With negative values, will move up (use move or move_up instead).

        Returns:
            int: The new position of the download.
        """
        return self.client.change_position(download.gid, pos, "POS_CUR")

    def move_to_top(self, download):
        """
        Move a download to the top of the queue.

        Args:
            download (:class:`~aria2p.downloads.Download`): the download object to move.

        Returns:
            int: The new position of the download.
        """
        return self.client.change_position(download.gid, 0, "POS_SET")

    def move_to_bottom(self, download):
        """
        Move a download to the bottom of the queue.

        Args:
            download (:class:`~aria2p.downloads.Download`): the download object to move.

        Returns:
            int: The new position of the download.
        """
        return self.client.change_position(download.gid, 0, "POS_END")

    def remove(self, downloads, force=False, files=False):
        """
        Remove the given downloads from the list.

        Args:
            downloads (list of :class:`~aria2p.downloads.Download`): the list of downloads to remove.
            force (bool): whether to force the removal or not.
            files (bool): whether to remove downloads files as well.

        Returns:
            list of bool: Success or failure of the operation for each given download.
        """
        # TODO: batch/multicall candidate
        if force:
            remove_func = self.client.force_remove
        else:
            remove_func = self.client.remove

        result = []

        for download in downloads:
            if download.is_complete or download.is_removed or download.has_failed:
                logger.debug(f"Try to remove download result {download.gid}")
                try:
                    self.client.remove_download_result(download.gid)
                except ClientException as error:
                    logger.exception(error)
                    result.append(error)
                else:
                    logger.success(f"Removed download result {download.gid}")
                    result.append(True)
            else:
                logger.debug(f"Try to remove download {download.gid}")
                try:
                    removed_gid = remove_func(download.gid)
                except ClientException as error:
                    logger.exception(error)
                    result.append(error)
                else:
                    logger.success(f"Removed download {download.gid}")
                    result.append(True)
                    try:
                        self.client.remove_download_result(download.gid)
                    except ClientException as error2:
                        logger.debug(f"Failed to remove download result {download.gid}")
                        logger.opt(exception=True).trace(error2)
                    if removed_gid != download.gid:
                        logger.debug(
                            f"Removed download GID#{removed_gid} is different than download GID#{download.gid}"
                        )
                        try:
                            self.client.remove_download_result(removed_gid)
                        except ClientException as error2:
                            logger.debug(f"Failed to remove download result {removed_gid}")
                            logger.opt(exception=True).trace(error2)
            if files and result[-1]:
                self.remove_files([download], force=force)

        return result

    def remove_all(self, force=False):
        """
        Remove all downloads from the list.

        Args:
            force (bool): whether to force the removal or not.

        Returns:
            bool: Success or failure of the operation to remove all downloads.
        """
        return all(self.remove(self.get_downloads(), force=force))

    def pause(self, downloads, force=False):
        """
        Remove the given downloads from the list.

        Args:
            downloads (list of :class:`~aria2p.downloads.Download`): the list of downloads to remove.
            force (bool): whether to pause immediately without contacting servers or not.

        Returns:
            list of bool: Success or failure of the operation for each given download.
        """
        # TODO: batch/multicall candidate
        if force:
            pause_func = self.client.force_pause
        else:
            pause_func = self.client.pause

        result = []

        for download in downloads:
            try:
                pause_func(download.gid)
            except ClientException as error:
                logger.debug(f"Failed to pause download {download.gid}")
                logger.opt(exception=True).trace(error)
                result.append(error)
            else:
                result.append(True)

        return result

    def pause_all(self, force=False):
        """
        Remove the given downloads from the list.

        Args:
            force (bool): whether to pause immediately without contacting servers or not.

        Returns:
            bool: Success or failure of the operation to pause all downloads.
        """
        if force:
            pause_func = self.client.force_pause_all
        else:
            pause_func = self.client.pause_all
        return pause_func() == "OK"

    def resume(self, downloads):
        """
        Resume (unpause) the given downloads.

        Args:
            downloads (list of :class:`~aria2p.downloads.Download`): the list of downloads to resume.

        Returns:
            list of bool: Success or failure of the operation for each given download.
        """
        # TODO: batch/multicall candidate
        result = []

        for download in downloads:
            try:
                self.client.unpause(download.gid)
            except ClientException as error:
                logger.debug(f"Failed to resume download {download.gid}")
                logger.opt(exception=True).trace(error)
                result.append(error)
            else:
                result.append(True)

        return result

    def resume_all(self):
        """
        Resume (unpause) all downloads.

        Returns:
            bool: Success or failure of the operation to resume all downloads.
        """
        return self.client.unpause_all() == "OK"

    def autopurge(self):
        """
        Purge completed, removed or failed downloads from the queue.

        Returns:
            bool: Success or failure of the operation.
        """
        version = get_version()
        if version.major == 0 and 9 > version.minor >= 7:
            logger.warning("Future change warning: API method 'autopurge' will be renamed 'purge' in version 0.9.0.")
        return self.client.purge_download_result()

    def purge(self, downloads):
        """
        Purge given downloads from the queue.

        Returns:
            list of bool: Success or failure of the operation for each download.
        """
        # TODO: batch/multicall candidate
        logger.warning(
            "Deprecation warning: API method 'purge' is deprecated in favor of method 'remove', "
            "and will be removed in version 0.7.0."
        )
        result = []

        for download in downloads:
            try:
                self.client.remove_download_result(download.gid)
            except ClientException as error:
                logger.exception(error)
                result.append(error)
            else:
                result.append(True)

        return result

    def purge_all(self):
        """
        Purge all downloads from the list.

        Returns:
            bool: Success or failure of the operation to purge all downloads.
        """
        logger.warning(
            "Deprecation warning: API method 'purge_all' is deprecated, and will be removed in version 0.7.0."
        )
        return all(self.purge(self.get_downloads()))

    def get_options(self, downloads):
        """
        Get options for each of the given downloads.

        Args:
            downloads (list of :class:`~aria2p.downloads.Download`): the list of downloads to get the options of.

        Returns:
            list of :class:`~aria2p.options.Options`: options object for each given download.
        """
        # TODO: batch/multicall candidate
        options = []
        for download in downloads:
            options.append(Options(self, self.client.get_option(download.gid), download))
        return options

    def get_global_options(self):
        """
        Get the global options.

        Returns:
            :class:`~aria2p.options.Options` instance: the global aria2c options.
        """
        return Options(self, self.client.get_global_option())

    def set_options(self, options, downloads):
        """
        Set options for specific downloads.

        Args:
            options (:class:`~aria2p.options.Options` or dict): an instance of the ``Options`` class or a dictionary
              containing aria2c options to create the download with.
            downloads (list of :class:`~aria2p.downloads.Download`): the list of downloads to set the options for.

        Returns:
            list of bool: Success or failure of the operation for changing options for each given download.
        """
        if isinstance(options, Options):
            client_options = options.get_struct()
        else:
            client_options = options

        # TODO: batch/multicall candidate
        results = []
        for download in downloads:
            results.append(self.client.change_option(download.gid, client_options) == "OK")
        return results

    def set_global_options(self, options):
        """
        Set global options.

        Args:
            options (:class:`~aria2p.options.Options` or dict): an instance of the ``Options`` class or a dictionary
              containing aria2c options to create the download with.

        Returns:
            bool: Success or failure of the operation for changing global options.
        """
        if isinstance(options, Options):
            client_options = options.get_struct()
        else:
            client_options = options

        return self.client.change_global_option(client_options) == "OK"

    def get_stats(self):
        """
        Get the stats of the remote aria2c process.

        Returns:
            :class:`~aria2p.stats.Stats` instance: the global stats returned by the remote process.
        """
        return Stats(self.client.get_global_stat())

    @staticmethod
    def remove_files(downloads, force=False):
        """
        Remove downloaded files.

        Args:
            downloads (list of :class:`~aria2p.downloads.Download`):  the list of downloads for which to remove files.
            force (bool): whether to remove files even if download is not complete.

        Returns:
            list of bool: Success or failure of the operation for each given download.
        """
        results = []
        for download in downloads:
            if download.is_complete or force:
                for path in download.root_files_paths:
                    if path.is_dir():
                        shutil.rmtree(str(path))
                    else:
                        path.unlink()
                results.append(True)
            else:
                results.append(False)
        return results

    @staticmethod
    def move_files(downloads, to_directory, force=False):
        """
        Move downloaded files to another directory.

        Args:
            downloads (list of :class:`~aria2p.downloads.Download`):  the list of downloads for which to move files.
            to_directory (str/Path): the target directory to move files to.
            force (bool): whether to move files even if download is not complete.

        Returns:
            list of bool: Success or failure of the operation for each given download.
        """
        if isinstance(to_directory, str):
            to_directory = Path(to_directory)

        # raises FileExistsError when target is already a file
        to_directory.mkdir(parents=True, exist_ok=True)

        results = []
        for download in downloads:
            if download.is_complete or force:
                for path in download.root_files_paths:
                    shutil.move(str(path), str(to_directory))
                results.append(True)
            else:
                results.append(False)
        return results

    @staticmethod
    def copy_files(downloads, to_directory, force=False):
        """
        Copy downloaded files to another directory.

        Args:
            downloads (list of :class:`~aria2p.downloads.Download`):  the list of downloads for which to move files.
            to_directory (str/Path): the target directory to copy files into.
            force (bool): whether to move files even if download is not complete.

        Returns:
            list of bool: Success or failure of the operation for each given download.
        """
        if isinstance(to_directory, str):
            to_directory = Path(to_directory)

        # raises FileExistsError when target is already a file
        to_directory.mkdir(parents=True, exist_ok=True)

        results = []
        for download in downloads:
            if download.is_complete or force:
                for path in download.root_files_paths:
                    if path.is_dir():
                        shutil.copytree(str(path), str(to_directory / path.name))
                    elif path.is_file():
                        shutil.copy(str(path), str(to_directory))

                results.append(True)
            else:
                results.append(False)
        return results

    def listen_to_notifications(self, threaded=False, **kwargs):
        """
        Start listening to aria2 notifications via WebSocket.

        This method differs from :meth:`~aria2p.client.Client.listen_to_notifications` in that it expects callbacks
        accepting two arguments, ``api`` and ``gid``, instead of only ``gid``.
        Accepting ``api`` allows to use the high-level methods of the :class:`~aria2p.api.API` class.

        Stop listening to notifications with the :meth:`~aria2p.api.API.stop_listening` method.

        Args:
            threaded (bool): Whether to start the listening loop in a thread or not (non-blocking or blocking).
        """

        def closure(callback):
            return (lambda gid: callback(self, gid)) if callable(callback) else None

        def run():
            self.client.listen_to_notifications(
                **{key: closure(value) if key.startswith("on_") else value for key, value in kwargs.items()}
            )

        if threaded:
            kwargs["handle_signals"] = False
            self.listener = threading.Thread(target=run)
            self.listener.start()
        else:
            run()

    def stop_listening(self):
        """
        Stop listening to notifications.

        If the listening loop was threaded, this method will wait for the thread to finish.
        The time it takes for the thread to finish will depend on the timeout given while calling
        :meth:`~aria2p.api.API.listen_to_notifications`.
        """
        self.client.stop_listening()
        if self.listener:
            self.listener.join()
        self.listener = None



"""
This module defines the Stats class, which holds information retrieved with the ``get_global_stat`` method of the
client.
"""
from .utils import human_readable_bytes


class Stats:
    """This class holds information retrieved with the ``get_global_stat`` method of the client."""

    def __init__(self, struct):
        """
        Initialization method.

        Args:
            struct (dict): a dictionary Python object returned by the JSON-RPC client.
        """
        self._struct = struct or {}

    @property
    def download_speed(self):
        """Overall download speed (byte/sec)."""
        return int(self._struct.get("downloadSpeed"))

    def download_speed_string(self, human_readable=True):
        """
        Return the download speed as string.

        Args:
            human_readable (bool): return in human readable format or not.

        Returns:
            str: the download speed string.
        """
        if human_readable:
            return human_readable_bytes(self.download_speed, delim=" ", postfix="/s")
        return str(self.download_speed) + " B/s"

    @property
    def upload_speed(self):
        """Overall upload speed(byte/sec)."""
        return int(self._struct.get("uploadSpeed"))

    def upload_speed_string(self, human_readable=True):
        """
        Return the upload speed as string.

        Args:
            human_readable (bool): return in human readable format or not.

        Returns:
            str: the upload speed string.
        """
        if human_readable:
            return human_readable_bytes(self.upload_speed, delim=" ", postfix="/s")
        return str(self.upload_speed) + " B/s"

    @property
    def num_active(self):
        """The number of active downloads."""
        return int(self._struct.get("numActive"))

    @property
    def num_waiting(self):
        """The number of waiting downloads."""
        return int(self._struct.get("numWaiting"))

    @property
    def num_stopped(self):
        """
        The number of stopped downloads in the current session. This value is capped by the --max-download-result
        option.
        """
        return int(self._struct.get("numStopped"))

    @property
    def num_stopped_total(self):
        """The number of stopped downloads in the current session and not capped by the --max-download-result option."""
        return int(self._struct.get("numStoppedTotal"))



"""
Module that contains the command line application.

Why does this file exist, and why not put this in __main__?

You might be tempted to import things from __main__ later,
but that will cause problems: the code will get executed twice:

- When you run `python -m aria2p` python will execute
  ``__main__.py`` as a script. That means there won't be any
  ``aria2p.__main__`` in ``sys.modules``.
- When you import __main__ it will get executed again (as a module) because
  there's no ``aria2p.__main__`` in ``sys.modules``.

Also see http://click.pocoo.org/5/setuptools/#setuptools-integration.
"""

import argparse
import importlib.util
import json
import sys
from pathlib import Path

import requests
from loguru import logger

from aria2p import Download, enable_logger

from .api import API
from .client import DEFAULT_HOST, DEFAULT_PORT, Client, ClientException
from .interface import Interface
from .utils import get_version


# ============ MAIN FUNCTION ============ #
def main(args=None):
    """The main function, which is executed when you type ``aria2p`` or ``python -m aria2p``."""

    parser = get_parser()
    args = parser.parse_args(args=args)
    kwargs = args.__dict__

    log_level = kwargs.pop("log_level")
    log_path = kwargs.pop("log_path")

    if log_path:
        log_path = Path(log_path)
        if log_path.is_dir():
            log_path = log_path / "aria2p-{time}.log"
        enable_logger(sink=log_path, level=log_level or "WARNING")
    elif log_level:
        enable_logger(sink=sys.stderr, level=log_level)

    logger.debug("Checking arguments")
    check_args(parser, args)

    logger.debug("Instantiating API")
    api = API(Client(host=kwargs.pop("host"), port=kwargs.pop("port"), secret=kwargs.pop("secret")))

    logger.info(f"API instantiated: {api!r}")

    # Warn if no aria2 daemon process seems to be running
    logger.debug("Testing connection")
    try:
        api.client.get_version()
    except requests.ConnectionError as error:
        print(f"[ERROR] {error}", file=sys.stderr)
        print(file=sys.stderr)
        print("Please make sure that an instance of aria2c is running with RPC mode enabled,", file=sys.stderr)
        print("and that you have provided the right host, port and secret token.", file=sys.stderr)
        print("More information at https://aria2p.readthedocs.io/en/latest.", file=sys.stderr)
        return 2

    subcommands = {
        None: subcommand_show,
        "show": subcommand_show,
        "top": subcommand_top,
        "call": subcommand_call,
        "add-magnet": subcommand_add_magnet,
        "add-torrent": subcommand_add_torrent,
        "add-metalink": subcommand_add_metalink,
        "pause": subcommand_pause,
        "stop": subcommand_pause,  # alias for pause
        "resume": subcommand_resume,
        "start": subcommand_resume,  # alias for resume
        "remove": subcommand_remove,
        "rm": subcommand_remove,  # alias for remove
        "del": subcommand_remove,  # alias for remove
        "delete": subcommand_remove,  # alias for remove
        "purge": subcommand_purge,
        "clear": subcommand_purge,  # alias for purge
        "autopurge": subcommand_autopurge,
        "autoclear": subcommand_autopurge,  # alias for autopurge
        "listen": subcommand_listen,
    }

    subcommand = kwargs.pop("subcommand")

    if subcommand:
        logger.debug("Running subcommand " + subcommand)
    else:
        print(
            "Future change warning: starting at version 0.6, the default command will be 'top' instead of 'show'.",
            file=sys.stderr,
        )
    try:
        return subcommands.get(subcommand)(api, **kwargs)
    except ClientException as error:
        print(error.message, file=sys.stderr)
        return error.code


def check_args(parser, args):
    """Additional checks for command line arguments."""
    subparser = [action for action in parser._actions if isinstance(action, argparse._SubParsersAction)][0].choices

    if args.subcommand in ("pause", "stop", "remove", "rm", "del", "delete", "resume", "start", "purge", "clear"):
        if not args.do_all and not args.gids:
            subparser[args.subcommand].error("the following arguments are required: gids or --all")
        elif args.do_all and args.gids:
            subparser[args.subcommand].error("argument -a/--all: not allowed with arguments gids")


def get_parser():
    """Return a parser for the command-line options and arguments."""
    usage = "%(prog)s [GLOBAL_OPTS...] COMMAND [COMMAND_OPTS...]"
    description = "Command-line tool and Python library to interact with an `aria2c` daemon process through JSON-RPC."
    parser = argparse.ArgumentParser(add_help=False, usage=usage, description=description, prog="aria2p")

    main_help = "Show this help message and exit. Commands also accept the -h/--help option."
    subcommand_help = "Show this help message and exit."

    global_options = parser.add_argument_group(title="Global options")
    global_options.add_argument("-h", "--help", action="help", help=main_help)

    global_options.add_argument(
        "-p", "--port", dest="port", default=DEFAULT_PORT, type=int, help="Port to use to connect to the remote server."
    )
    global_options.add_argument(
        "-H", "--host", dest="host", default=DEFAULT_HOST, help="Host address for the remote server."
    )
    global_options.add_argument(
        "-s", "--secret", dest="secret", default="", help="Secret token to use to connect to the remote server."
    )
    global_options.add_argument(
        "-L",
        "--log-level",
        dest="log_level",
        default=None,
        help="Log level to use",
        choices=("TRACE", "DEBUG", "INFO", "SUCCESS", "WARNING", "ERROR", "CRITICAL"),
        type=str.upper,
    )
    global_options.add_argument(
        "-P", "--log-path", dest="log_path", default=None, help="Log path to use. Can be a directory or a file."
    )

    # ========= SUBPARSERS ========= #
    subparsers = parser.add_subparsers(dest="subcommand", title="Commands", metavar="", prog="aria2p")

    def subparser(command, text, **kwargs):
        sub = subparsers.add_parser(command, add_help=False, help=text, description=text, **kwargs)
        sub.add_argument("-h", "--help", action="help", help=subcommand_help)
        return sub

    add_magnet_parser = subparser("add-magnet", "Add a download with a Magnet URI.")
    add_metalink_parser = subparser("add-metalink", "Add a download with a Metalink file.")
    add_torrent_parser = subparser("add-torrent", "Add a download with a torrent file.")
    subparser("autopurge", "Automatically purge completed/removed/failed downloads.", aliases=["autoclear"])
    call_parser = subparser("call", "Call a remote method through the JSON-RPC client.")
    pause_parser = subparser("pause", "Pause downloads.", aliases=["stop"])
    pause_all_parser = subparser("pause-all", "Pause all downloads.")
    purge_parser = subparser("purge", "Purge downloads.", aliases=["clear"])
    remove_parser = subparser("remove", "Remove downloads.", aliases=["rm", "del", "delete"])
    remove_all_parser = subparser("remove-all", "Remove all downloads.")
    resume_parser = subparser("resume", "Resume downloads.", aliases=["start"])
    subparser("resume-all", "Resume all downloads.")
    subparser("show", "Show the download progression.")
    subparser("top", "Launch the top-like interactive interface.")
    listen_parser = subparser("listen", "Listen to notifications.")

    # ========= CALL PARSER ========= #
    call_parser.add_argument(
        "method",
        help=(
            "The method to call (case insensitive). "
            "Dashes and underscores will be removed so you can use as many as you want, or none. "
            "Prefixes like 'aria2.' or 'system.' are also optional."
        ),
    )
    call_parser_mxg = call_parser.add_mutually_exclusive_group()
    call_parser_mxg.add_argument(
        "-P", "--params-list", dest="params", nargs="+", help="Parameters as a list of strings."
    )
    call_parser_mxg.add_argument(
        "-J",
        "--json-params",
        dest="params",
        help="Parameters as a JSON string. You should always wrap it at least once in an array '[]'.",
    )

    # ========= ADD MAGNET PARSER ========= #
    add_magnet_parser.add_argument("uri", help="The magnet URI to use.")

    # ========= ADD TORRENT PARSER ========= #
    add_torrent_parser.add_argument("torrent_file", help="The path to the torrent file.")

    # ========= ADD METALINK PARSER ========= #
    add_metalink_parser.add_argument("metalink_file", help="The path to the metalink file.")

    # ========= PAUSE PARSER ========= #
    pause_parser.add_argument("gids", nargs="*", help="The GIDs of the downloads to pause.")
    pause_parser.add_argument("-a", "--all", action="store_true", dest="do_all", help="Pause all the downloads.")
    pause_parser.add_argument(
        "-f", "--force", dest="force", action="store_true", help="Pause without contacting servers first."
    )

    # ========= PAUSE ALL PARSER ========= #
    pause_all_parser.add_argument(
        "-f", "--force", dest="force", action="store_true", help="Pause without contacting servers first."
    )

    # ========= RESUME PARSER ========= #
    resume_parser.add_argument("gids", nargs="*", help="The GIDs of the downloads to resume.")
    resume_parser.add_argument("-a", "--all", action="store_true", dest="do_all", help="Resume all the downloads.")

    # ========= REMOVE PARSER ========= #
    remove_parser.add_argument("gids", nargs="*", help="The GIDs of the downloads to remove.")
    remove_parser.add_argument("-a", "--all", action="store_true", dest="do_all", help="Remove all the downloads.")
    remove_parser.add_argument(
        "-f", "--force", dest="force", action="store_true", help="Remove without contacting servers first."
    )

    # ========= REMOVE ALL PARSER ========= #
    remove_all_parser.add_argument(
        "-f", "--force", dest="force", action="store_true", help="Remove without contacting servers first."
    )

    # ========= PURGE PARSER ========= #
    purge_parser.add_argument("gids", nargs="*", help="The GIDs of the downloads to purge.")
    purge_parser.add_argument("-a", "--all", action="store_true", dest="do_all", help="Purge all the downloads.")

    # ========= LISTEN PARSER ========= #
    listen_parser.add_argument(
        "-c",
        "--callbacks-module",
        dest="callbacks_module",
        help="Path to the Python module defining your notifications callbacks.",
    )
    listen_parser.add_argument(
        "event_types",
        nargs="*",
        help="The types of notifications to process: "
        "start, pause, stop, error, complete or btcomplete. "
        "Example: aria2p listen error btcomplete. "
        "Useful if you want to spawn multiple specialized aria2p listener, "
        "for example one for each type of notification, "
        "but still want to use only one callback file.",
    )
    listen_parser.add_argument(
        "-t",
        "--timeout",
        dest="timeout",
        type=float,
        default=5,
        help="Timeout in seconds to use when waiting for data over the WebSocket at each iteration. "
        "Use small values for faster reactivity when stopping to listen.",
    )

    return parser


# ============ SUBCOMMANDS ============ #
def subcommand_show(api):
    """
    Show subcommand.

    Args:
        api (API): the API instance to use.

    Returns:
        int: always 0.
    """
    downloads = api.get_downloads()

    print(
        f"{'GID':<17} "
        f"{'STATUS':<9} "
        f"{'PROGRESS':>8} "
        f"{'DOWN_SPEED':>12} "
        f"{'UP_SPEED':>12} "
        f"{'ETA':>8}  "
        f"NAME"
    )

    for download in downloads:
        print(
            f"{download.gid:<17} "
            f"{download.status:<9} "
            f"{download.progress_string():>8} "
            f"{download.download_speed_string():>12} "
            f"{download.upload_speed_string():>12} "
            f"{download.eta_string():>8}  "
            f"{download.name}"
        )

    return 0


def subcommand_top(api):
    """
    Top subcommand.

    Args:
        api (API): the API instance to use.

    Returns:
        int: always 0.
    """
    interface = Interface(api)
    success = interface.run()
    return 0 if success else 1


def subcommand_call(api, method, params):
    """
    Call subcommand.

    Args:
        api (API): the API instance to use.
        method (str): name of the method to call.
        params (str / list of str): parameters to use when calling method.

    Returns:
        int: always 0.
    """
    real_method = get_method(method)
    if real_method is None:
        print(f"aria2p: call: Unknown method {method}.", file=sys.stderr)
        print("  Run 'aria2p call listmethods' to list the available methods.", file=sys.stderr)
        return 1

    if isinstance(params, str):
        params = json.loads(params)
    elif params is None:
        params = []

    response = api.client.call(real_method, params)
    print(json.dumps(response))

    return 0


def get_method(name, default=None):
    """Return the actual aria2 method name from a differently formatted name."""
    methods = {}
    for method in Client.METHODS:
        methods[method.lower()] = method
        methods[method.split(".")[1].lower()] = method
    name = name.lower()
    name = name.replace("-", "")
    name = name.replace("_", "")
    return methods.get(name, default)


def subcommand_add_magnet(api, uri):
    """
    Add magnet subcommand.

    Args:
        api (API): the API instance to use.
        uri (str): the URI of the magnet.

    Returns:
        int: always 0.
    """
    new_download = api.add_magnet(uri)
    print(f"Created download {new_download.gid}")
    return 0


def subcommand_add_torrent(api, torrent_file):
    """
    Add torrent subcommand.

    Args:
        api (API): the API instance to use.
        torrent_file (str): the path to the torrent file.

    Returns:
        int: always 0.
    """
    new_download = api.add_torrent(torrent_file)
    print(f"Created download {new_download.gid}")
    return 0


def subcommand_add_metalink(api: API, metalink_file):
    """
    Add metalink subcommand.

    Args:
        api (API): the API instance to use.
        metalink_file (str): the path to the metalink file.

    Returns:
        int: always 0.
    """
    new_downloads = api.add_metalink(metalink_file)
    for download in new_downloads:
        print(f"Created download {download.gid}")
    return 0


def subcommand_pause(api: API, gids=None, do_all=False, force=False):
    """
    Pause subcommand.

    Args:
        api (API): the API instance to use.
        gids (list of str): the GIDs of the downloads to pause.
        do_all (bool): pause all downloads if True.
        force (bool): force pause or not (see API.pause).

    Returns:
        int: 0 if all success, 1 if one failure.
    """
    if do_all:
        if api.pause_all(force=force):
            return 0
        return 1

    downloads = [Download(api, {"gid": gid}) for gid in gids]
    result = api.pause(downloads, force=force)
    if all(result):
        return 0
    for item in result:
        if isinstance(item, ClientException):
            print(item, file=sys.stderr)
    return 1


def subcommand_resume(api: API, gids=None, do_all=False):
    """
    Resume subcommand.

    Args:
        api (API): the API instance to use.
        gids (list of str): the GIDs of the downloads to resume.
        do_all (bool): pause all downloads if True.

    Returns:
        int: 0 if all success, 1 if one failure.
    """
    if do_all:
        if api.resume_all():
            return 0
        return 1

    downloads = [Download(api, {"gid": gid}) for gid in gids]
    result = api.resume(downloads)
    if all(result):
        return 0
    for item in result:
        if isinstance(item, ClientException):
            print(item, file=sys.stderr)
    return 1


def subcommand_remove(api: API, gids=None, do_all=False, force=False):
    """
    Remove subcommand.

    Args:
        api (API): the API instance to use.
        gids (list of str): the GIDs of the downloads to remove.
        do_all (bool): pause all downloads if True.
        force (bool): force pause or not (see API.remove).

    Returns:
        int: 0 if all success, 1 if one failure.
    """
    if do_all:
        if api.remove_all():
            return 0
        return 1

    downloads = [Download(api, {"gid": gid}) for gid in gids]
    result = api.remove(downloads, force=force)
    if all(result):
        return 0
    for item in result:
        if isinstance(item, ClientException):
            print(item, file=sys.stderr)
    return 1


def subcommand_purge(api: API, gids=None, do_all=False):
    """
    Purge subcommand.

    Args:
        api (API): the API instance to use.
        gids (list of str): the GIDs of the downloads to purge.
        do_all (bool): pause all downloads if True.

    Returns:
        int: 0 if all success, 1 if one failure.
    """
    print(
        "Deprecation warning: command 'purge' is deprecated in favor of command 'remove', "
        "and will be removed in version 0.7.0.",
        file=sys.stderr,
    )
    if do_all:
        if api.purge_all():
            return 0
        return 1

    downloads = [Download(api, {"gid": gid}) for gid in gids]
    result = api.purge(downloads)
    if all(result):
        return 0
    for item in result:
        if isinstance(item, ClientException):
            print(item, file=sys.stderr)
    return 1


def subcommand_autopurge(api: API):
    """
    Autopurge subcommand.

    Args:
        api (API): the API instance to use.

    Returns:
        int: 0 if all success, 1 if one failure.
    """
    version = get_version()
    if version.major == 0 and 9 > version.minor >= 7:
        print(
            "Future change warning: command 'autopurge' will be renamed 'purge' in version 0.9.0, "
            "with an 'autoremove' alias.",
            file=sys.stderr,
        )
    if api.autopurge():
        return 0
    return 1


def subcommand_listen(api: API, callbacks_module=None, event_types=None, timeout=5):
    """
    Listen subcommand.

    Args:
        api (API): the API instance to use.
        callbacks_module (Path/str): the path to the module to import, containing the callbacks as functions.
        event_types (list of str): the event types to process.
        timeout (float/int): the timeout to pass to the WebSocket connection, in seconds.

    Returns:
        int: always 0.
    """
    if not callbacks_module:
        print("aria2p: listen: Please provide the callback module file path with -c option", file=sys.stderr)
        return 1

    if isinstance(callbacks_module, Path):
        callbacks_module = str(callbacks_module)

    if not event_types:
        event_types = ["start", "pause", "stop", "error", "complete", "btcomplete"]

    spec = importlib.util.spec_from_file_location("aria2p_callbacks", callbacks_module)
    callbacks = importlib.util.module_from_spec(spec)

    if callbacks is None:
        print(f"aria2p: Could not import module file {callbacks_module}", file=sys.stderr)
        return 1

    spec.loader.exec_module(callbacks)

    callbacks_kwargs = {}
    for callback_name in (
        "on_download_start",
        "on_download_pause",
        "on_download_stop",
        "on_download_error",
        "on_download_complete",
        "on_bt_download_complete",
    ):
        if callback_name[3:].replace("download", "").replace("_", "") in event_types:
            callback = getattr(callbacks, callback_name, None)
            if callback:
                callbacks_kwargs[callback_name] = callback

    api.listen_to_notifications(timeout=timeout, handle_signals=True, threaded=False, **callbacks_kwargs)
    return 0



"""
aria2p package.

This package provides a command-line tool and a Python library to interact
with an ``aria2c`` daemon process through JSON-RPC.

If you read this message, you probably want to learn about the library and not the command-line tool:
please refer to the README.md included in this package to get the link to the official documentation.
"""

import sys

from loguru import logger

from .api import API
from .client import Client, ClientException
from .downloads import BitTorrent, Download, File
from .options import Options
from .stats import Stats

logger.disable("aria2p")


def enable_logger(sink=sys.stderr, level="WARNING"):
    """
    Enable the logging of messages.

    Configure the ``logger`` variable imported from ``loguru``.

    Args:
        sink (file): An opened file pointer, or stream handler. Default to standard error.
        level (str): The log level to use. Possible values are TRACE, DEBUG, INFO, WARNING, ERROR, CRITICAL.
            Default to WARNING.
    """
    logger.remove()
    logger.configure(handlers=[{"sink": sink, "level": level}])
    logger.enable("aria2p")


__all__ = ["API", "ClientException", "Client", "Download", "BitTorrent", "File", "Options", "Stats", "enable_logger"]
